'Middleware for HFS 2'

import argparse
import urllib.parse
from http.client import HTTPConnection
from http.server import ThreadingHTTPServer, BaseHTTPRequestHandler

def match_address(address: str, pattern: str) -> bool:
    'Match address with pattern'
    if '.' in address:
        return all([
            i == j or j == '*'
            for i, j in zip(
                address.split('.'),
                pattern.split('.')
            )
        ])
    elif ':' in address:
        # sorry, TODO
        return True
    else:
        return False

class MiddlewareHandler(BaseHTTPRequestHandler):
    'Middleware Handler'

    buffer = 4 * 1024 * 1024
    'buffer size to read/write a request'
    server = None

    def send_response(self, code, message=None):
        self.log_request(code)
        self.send_response_only(code, message)

    def do_whatever(self):
        'Just proxy through'
        server = self.server
        client_address = self.headers.get('X-Forwarded-For', self.client_address[0])
        # find if banned
        if client_address in server.ban_list:
            return
        for addr in server.ban_list_mask:
            for i in addr:
                if match_address(client_address, i):
                    return
        host = self.headers.get('Host')
        # yes, two checks required, if considering port is included
        if not host:
            server.ban(client_address, 'Request without Host header')
        if ':' in host:
            host = host.split(':')[0]
        if not host:
            server.ban(client_address, 'Request without Host header')
        path = self.path
        # for better accuracy identifing
        upath = urllib.parse.unquote(path)
        if '/..' in upath or '../' in upath or 'C:' in upath or '\\' in path:
            server.ban(client_address, 'Attempting dangerous access')
        if '?=PHP' in upath or '<?php' in upath:
            server.ban(client_address, 'Attempting PHP attack')
        if 'SELECT ' in upath:
            server.ban(client_address, 'Attempting SQL attack')
        if ('alert(' in upath or 'toString(' in upath or '<script' in upath
            or 'window.' in upath or 'document.' in upath or 'onload=' in upath):
            server.ban(client_address, 'Attempting XSS attack')
        if '{.' in upath or '.}' in upath or '\0' in upath:
            server.ban(client_address, 'Attempting macro attack')
        # Virtual Host
        if host in server.virtual_hosts and not path.startswith('/~'):
            if not path.startswith(server.virtual_hosts[host]):
                path = server.virtual_hosts[host] + path
        connection = HTTPConnection(server.hfs_bind, server.hfs_port)
        connection.connect()
        connection.putrequest(self.command, path)
        for i in self.headers:
            connection.putheader(i, self.headers[i])
        connection.putheader('X-Forwarded-For', client_address)
        connection.endheaders()
        if self.command == 'POST':
            if self.path == '/' and 'upload' not in path:
                server.ban(client_address, 'Attempting POST in root directory')
            content_length = self.headers.get('Content-Length')
            if content_length is not None and content_length.isdecimal():
                content_length = int(content_length)
                while data := self.rfile.read(min(self.buffer, content_length)):
                    if b'<?php\n' in data:
                        server.ban(client_address, 'Upload of PHP script')
                    if b'{.' in data and b'.}' in data:
                        server.ban(client_address, 'Upload of macro')
                    connection.send(data)
                    content_length -= len(data)
        response = connection.getresponse()
        self.send_response(response.code)
        for i in response.headers:
            if i in server.header_blacklist:
                continue
            self.send_header(i, response.headers[i])
        self.end_headers()
        if self.command == 'HEAD':
            return
        while data := response.read(self.buffer):
            self.wfile.write(data)
        self.wfile.flush()

    do_GET = do_whatever
    do_POST = do_whatever
    do_HEAD = do_whatever

class MiddlewareServer(ThreadingHTTPServer):
    'Middleware Server'

    hfs_port = 8080
    hfs_bind = '127.0.0.1'
    ban_list = []
    ban_list_mask = []
    header_blacklist = [
        'Server'
    ]
    virtual_hosts = {}

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        with open('banlist.txt', 'a', encoding='utf-8') as file:
            # for creating it without `import os` if doesn't exist yet
            pass
        with open('banlist.txt', 'r', encoding='utf-8') as file:
            while line := file.readline():
                if line.strip() == '' or line[0] == '#':
                    continue
                line = line[0:-1]   # remove \n
                if '*' in line:
                    self.ban_list_mask.append(line)
                else:
                    self.ban_list.append(line)

    def ban(self, address, reason=None):
        'Ban an address, optional with a reason'
        print(f'Banning {address} for {reason}')
        if '*' in address:
            self.ban_list_mask.append(address)
        else:
            self.ban_list.append(address)
        with open('banlist.txt', 'a', encoding='utf-8') as file:
            if reason is not None:
                file.write(f'# {reason}\n')
            file.write(address + '\n')
        # quit this thread and don't reply further
        # should only use in Threading server
        exit(0)

    def finish_request(self, request, client_address):
        handler = MiddlewareHandler(request, client_address, self)
        handler.server = self

def main():
    'Main routine for direct command line execution'
    parser = argparse.ArgumentParser(
        description='Middleware for HFS 2, providing some other goodies'
    )
    parser.add_argument('-b', '--bind', default='',
                        help='IP address where this middleware should bind to')
    parser.add_argument('-d', '--hfs-bind', default='127.0.0.1',
                        help='IP address where HFS is listening')
    parser.add_argument('-p', '--port', type=int, default=8081,
                        help='Port of this middleware')
    parser.add_argument('-q', '--hfs-port', type=int, default=8080,
                        help='Port of HFS')
    parser.add_argument('-v', '--virtual', required=False,
                        nargs='+', metavar='[domain[,path]]',
                        help='Virtual Host')
    args = parser.parse_args()
    server = MiddlewareServer((args.bind, args.port), MiddlewareHandler)
    server.hfs_port = args.hfs_port
    server.hfs_bind = args.hfs_bind
    if args.virtual:
        for item in args.virtual:
            i, j = item.split(',')
            server.virtual_hosts[i] = j
    try:
        print('http://%s:%i/' % (args.bind or 'localhost', args.port))
        server.serve_forever()
    except KeyboardInterrupt:
        server.server_close()

if __name__ == '__main__':
    main()
